/* GStreamer Wayland video sink
 *
 * Copyright (C) 2011 Intel Corporation
 * Copyright (C) 2011 Sreerenj Balachandran <sreerenj.balachandran@intel.com>
 * Copyright (C) 2012 Wim Taymans <wim.taymans@gmail.com>
 * Copyright (C) 2014 Collabora Ltd.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA.
 */

/**
 * SECTION:element-waylandsink
 * @title: waylandsink
 *
 *  The waylandsink is creating its own window and render the decoded video frames to that.
 *  Setup the Wayland environment as described in
 *  [Wayland](http://wayland.freedesktop.org/building.html) home page.
 *
 *  The current implementation is based on weston compositor.
 *
 * ## Example pipelines
 * |[
 * gst-launch-1.0 -v videotestsrc ! waylandsink
 * ]| test the video rendering in wayland
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "gstwaylandsink.h"
#include "wlvideoformat.h"
#include "wlbuffer.h"
#include "wlshmallocator.h"
#include "wllinuxdmabuf.h"
#include "wlgbmbufferbackend.h"

#include <gst/wayland/wayland.h>
#include <gst/video/videooverlay.h>

/* signals */
enum
{
  SIGNAL_0,
  LAST_SIGNAL
};

/* Properties */
enum
{
  PROP_0,
  PROP_DISPLAY,
  PROP_FULLSCREEN
};

GST_DEBUG_CATEGORY (gstwayland_debug);
#define GST_CAT_DEFAULT gstwayland_debug

#define WL_VIDEO_FORMATS \
    "{ BGRx, BGRA, RGBx, xBGR, xRGB, RGBA, ABGR, ARGB, RGB, BGR, " \
    "RGB16, BGR16, YUY2, YVYU, UYVY, AYUV, NV12, NV21, NV16, NV61, " \
    "YUV9, YVU9, Y41B, I420, YV12, Y42B, v308 }"

#define MAX_OVERLAY_WIDTH 1280
#define MAX_OVERLAY_HEIGHT 720

static GstStaticPadTemplate sink_template = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_VIDEO_CAPS_MAKE (WL_VIDEO_FORMATS) ";"
        "video/x-raw(" GST_CAPS_FEATURE_MEMORY_DMABUF ")")
    );

static void gst_wayland_sink_get_property (GObject * object,
    guint prop_id, GValue * value, GParamSpec * pspec);
static void gst_wayland_sink_set_property (GObject * object,
    guint prop_id, const GValue * value, GParamSpec * pspec);
static void gst_wayland_sink_finalize (GObject * object);

static GstStateChangeReturn gst_wayland_sink_change_state (GstElement * element,
    GstStateChange transition);
static void gst_wayland_sink_set_context (GstElement * element,
    GstContext * context);

static GstCaps *gst_wayland_sink_get_caps (GstBaseSink * bsink,
    GstCaps * filter);
static gboolean gst_wayland_sink_set_caps (GstBaseSink * bsink, GstCaps * caps);
static gboolean
gst_wayland_sink_propose_allocation (GstBaseSink * bsink, GstQuery * query);
static GstFlowReturn gst_wayland_sink_show_frame (GstVideoSink * bsink,
    GstBuffer * buffer);

/* VideoOverlay interface */
static void gst_wayland_sink_videooverlay_init (GstVideoOverlayInterface *
    iface);
static void gst_wayland_sink_set_window_handle (GstVideoOverlay * overlay,
    guintptr handle);
static void gst_wayland_sink_set_render_rectangle (GstVideoOverlay * overlay,
    gint x, gint y, gint w, gint h);
static void gst_wayland_sink_expose (GstVideoOverlay * overlay);

/* WaylandVideo interface */
static void gst_wayland_sink_waylandvideo_init (GstWaylandVideoInterface *
    iface);
static void gst_wayland_sink_begin_geometry_change (GstWaylandVideo * video);
static void gst_wayland_sink_end_geometry_change (GstWaylandVideo * video);

#define gst_wayland_sink_parent_class parent_class
G_DEFINE_TYPE_WITH_CODE (GstWaylandSink, gst_wayland_sink, GST_TYPE_VIDEO_SINK,
    G_IMPLEMENT_INTERFACE (GST_TYPE_VIDEO_OVERLAY,
        gst_wayland_sink_videooverlay_init)
    G_IMPLEMENT_INTERFACE (GST_TYPE_WAYLAND_VIDEO,
        gst_wayland_sink_waylandvideo_init));
GST_ELEMENT_REGISTER_DEFINE (waylandsink, "waylandsink", GST_RANK_MARGINAL,
    GST_TYPE_WAYLAND_SINK);

/* A tiny GstVideoBufferPool subclass that modify the options to remove
 * VideoAlignment. To support VideoAlignment we would need to pass the padded
 * width/height + stride and use the viewporter interface to crop, a bit like
 * we use to do with XV. It would still be quite limited. It's a bit retro,
 * hopefully there will be a better Wayland interface in the future. */

GType gst_wayland_pool_get_type (void);

typedef struct
{
  GstVideoBufferPool parent;
} GstWaylandPool;

typedef struct
{
  GstVideoBufferPoolClass parent;
} GstWaylandPoolClass;

G_DEFINE_TYPE (GstWaylandPool, gst_wayland_pool, GST_TYPE_VIDEO_BUFFER_POOL);

static const gchar **
gst_wayland_pool_get_options (GstBufferPool * pool)
{
  static const gchar *options[] = { GST_BUFFER_POOL_OPTION_VIDEO_META, NULL };
  return options;
}

static void
gst_wayland_pool_class_init (GstWaylandPoolClass * klass)
{
  GstBufferPoolClass *pool_class = GST_BUFFER_POOL_CLASS (klass);
  pool_class->get_options = gst_wayland_pool_get_options;
}

static void
gst_wayland_pool_init (GstWaylandPool * pool)
{
}

static void
gst_wayland_sink_class_init (GstWaylandSinkClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;
  GstBaseSinkClass *gstbasesink_class;
  GstVideoSinkClass *gstvideosink_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;
  gstbasesink_class = (GstBaseSinkClass *) klass;
  gstvideosink_class = (GstVideoSinkClass *) klass;

  gobject_class->set_property = gst_wayland_sink_set_property;
  gobject_class->get_property = gst_wayland_sink_get_property;
  gobject_class->finalize = GST_DEBUG_FUNCPTR (gst_wayland_sink_finalize);

  gst_element_class_add_static_pad_template (gstelement_class, &sink_template);

  gst_element_class_set_static_metadata (gstelement_class,
      "wayland video sink", "Sink/Video",
      "Output to wayland surface",
      "Sreerenj Balachandran <sreerenj.balachandran@intel.com>, "
      "George Kiagiadakis <george.kiagiadakis@collabora.com>");

  gstelement_class->change_state =
      GST_DEBUG_FUNCPTR (gst_wayland_sink_change_state);
  gstelement_class->set_context =
      GST_DEBUG_FUNCPTR (gst_wayland_sink_set_context);

  gstbasesink_class->get_caps = GST_DEBUG_FUNCPTR (gst_wayland_sink_get_caps);
  gstbasesink_class->set_caps = GST_DEBUG_FUNCPTR (gst_wayland_sink_set_caps);
  gstbasesink_class->propose_allocation =
      GST_DEBUG_FUNCPTR (gst_wayland_sink_propose_allocation);

  gstvideosink_class->show_frame =
      GST_DEBUG_FUNCPTR (gst_wayland_sink_show_frame);

  g_object_class_install_property (gobject_class, PROP_DISPLAY,
      g_param_spec_string ("display", "Wayland Display name", "Wayland "
          "display name to connect to, if not supplied via the GstContext",
          NULL, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  g_object_class_install_property (gobject_class, PROP_FULLSCREEN,
      g_param_spec_boolean ("fullscreen", "Fullscreen",
          "Whether the surface should be made fullscreen ", FALSE,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  gst_type_mark_as_plugin_api (GST_TYPE_WAYLAND_VIDEO, 0);
}

static void
gst_wayland_sink_init (GstWaylandSink * sink)
{
  g_mutex_init (&sink->display_lock);
  g_mutex_init (&sink->render_lock);
}

static void
gst_wayland_sink_set_fullscreen (GstWaylandSink * sink, gboolean fullscreen)
{
  if (fullscreen == sink->fullscreen)
    return;

  g_mutex_lock (&sink->render_lock);
  sink->fullscreen = fullscreen;
  gst_wl_window_ensure_fullscreen (sink->window, fullscreen);
  g_mutex_unlock (&sink->render_lock);
}

static void
gst_wayland_sink_get_property (GObject * object,
    guint prop_id, GValue * value, GParamSpec * pspec)
{
  GstWaylandSink *sink = GST_WAYLAND_SINK (object);

  switch (prop_id) {
    case PROP_DISPLAY:
      GST_OBJECT_LOCK (sink);
      g_value_set_string (value, sink->display_name);
      GST_OBJECT_UNLOCK (sink);
      break;
    case PROP_FULLSCREEN:
      GST_OBJECT_LOCK (sink);
      g_value_set_boolean (value, sink->fullscreen);
      GST_OBJECT_UNLOCK (sink);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_wayland_sink_set_property (GObject * object,
    guint prop_id, const GValue * value, GParamSpec * pspec)
{
  GstWaylandSink *sink = GST_WAYLAND_SINK (object);

  switch (prop_id) {
    case PROP_DISPLAY:
      GST_OBJECT_LOCK (sink);
      sink->display_name = g_value_dup_string (value);
      GST_OBJECT_UNLOCK (sink);
      break;
    case PROP_FULLSCREEN:
      GST_OBJECT_LOCK (sink);
      gst_wayland_sink_set_fullscreen (sink, g_value_get_boolean (value));
      GST_OBJECT_UNLOCK (sink);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_wayland_sink_finalize (GObject * object)
{
  GstWaylandSink *sink = GST_WAYLAND_SINK (object);

  GST_DEBUG_OBJECT (sink, "Finalizing the sink..");

  if (sink->last_buffer)
    gst_buffer_unref (sink->last_buffer);
  if (sink->display)
    g_object_unref (sink->display);
  if (sink->window)
    g_object_unref (sink->window);
  if (sink->pool)
    gst_object_unref (sink->pool);

  g_free (sink->display_name);

  g_mutex_clear (&sink->display_lock);
  g_mutex_clear (&sink->render_lock);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

/* must be called with the display_lock */
static void
gst_wayland_sink_set_display_from_context (GstWaylandSink * sink,
    GstContext * context)
{
  struct wl_display *display;
  GError *error = NULL;

  display = gst_wayland_display_handle_context_get_handle (context);
  sink->display = gst_wl_display_new_existing (display, FALSE, &error);

  if (error) {
    GST_ELEMENT_WARNING (sink, RESOURCE, OPEN_READ_WRITE,
        ("Could not set display handle"),
        ("Failed to use the external wayland display: '%s'", error->message));
    g_error_free (error);
  }
}

static gboolean
gst_wayland_sink_find_display (GstWaylandSink * sink)
{
  GstQuery *query;
  GstMessage *msg;
  GstContext *context = NULL;
  GError *error = NULL;
  gboolean ret = TRUE;

  g_mutex_lock (&sink->display_lock);

  if (!sink->display) {
    /* first query upstream for the needed display handle */
    query = gst_query_new_context (GST_WAYLAND_DISPLAY_HANDLE_CONTEXT_TYPE);
    if (gst_pad_peer_query (GST_VIDEO_SINK_PAD (sink), query)) {
      gst_query_parse_context (query, &context);
      gst_wayland_sink_set_display_from_context (sink, context);
    }
    gst_query_unref (query);

    if (G_LIKELY (!sink->display)) {
      /* now ask the application to set the display handle */
      msg = gst_message_new_need_context (GST_OBJECT_CAST (sink),
          GST_WAYLAND_DISPLAY_HANDLE_CONTEXT_TYPE);

      g_mutex_unlock (&sink->display_lock);
      gst_element_post_message (GST_ELEMENT_CAST (sink), msg);
      /* at this point we expect gst_wayland_sink_set_context
       * to get called and fill sink->display */
      g_mutex_lock (&sink->display_lock);

      if (!sink->display) {
        /* if the application didn't set a display, let's create it ourselves */
        GST_OBJECT_LOCK (sink);
        sink->display = gst_wl_display_new (sink->display_name, &error);
        GST_OBJECT_UNLOCK (sink);

        if (error) {
          GST_ELEMENT_WARNING (sink, RESOURCE, OPEN_READ_WRITE,
              ("Could not initialise Wayland output"),
              ("Failed to create GstWlDisplay: '%s'", error->message));
          g_error_free (error);
          ret = FALSE;
        }
      }
    }
  }

  g_mutex_unlock (&sink->display_lock);

  return ret;
}

static void
overlay_shutdown (GstWaylandSink * sink)
{
  if (sink->overlay_pool) {
    gst_object_unref (sink->overlay_pool);
    sink->overlay_pool = NULL;
  }

  if (sink->spu_subsurface) {
    wl_subsurface_destroy (sink->spu_subsurface);
    sink->spu_subsurface = NULL;
  }

  if (sink->spu_viewport) {
    wp_viewport_destroy (sink->spu_viewport);
    sink->spu_viewport = NULL;
  }

  if (sink->spu_surface) {
    wl_surface_destroy (sink->spu_surface);
    sink->spu_surface = NULL;
  }
}

static GstStateChangeReturn
gst_wayland_sink_change_state (GstElement * element, GstStateChange transition)
{
  GstWaylandSink *sink = GST_WAYLAND_SINK (element);
  GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;

  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
      if (!gst_wayland_sink_find_display (sink))
        return GST_STATE_CHANGE_FAILURE;
      break;
    default:
      break;
  }

  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
  if (ret == GST_STATE_CHANGE_FAILURE)
    return ret;

  switch (transition) {
    case GST_STATE_CHANGE_PAUSED_TO_READY:
      gst_buffer_replace (&sink->last_buffer, NULL);
      overlay_shutdown (sink);
      if (sink->window) {
        if (gst_wl_window_is_toplevel (sink->window)) {
          g_clear_object (&sink->window);
        } else {
          /* remove buffer from surface, show nothing */
          gst_wl_window_render (sink->window, NULL, NULL, FALSE);
        }
      }

      g_mutex_lock (&sink->render_lock);
      if (sink->callback) {
        wl_callback_destroy (sink->callback);
        sink->callback = NULL;
      }
      sink->redraw_pending = FALSE;
      g_mutex_unlock (&sink->render_lock);
      break;
    case GST_STATE_CHANGE_READY_TO_NULL:
      g_mutex_lock (&sink->display_lock);
      /* If we had a toplevel window, we most likely have our own connection
       * to the display too, and it is a good idea to disconnect and allow
       * potentially the application to embed us with GstVideoOverlay
       * (which requires to re-use the same display connection as the parent
       * surface). If we didn't have a toplevel window, then the display
       * connection that we have is definitely shared with the application
       * and it's better to keep it around (together with the window handle)
       * to avoid requesting them again from the application if/when we are
       * restarted (GstVideoOverlay behaves like that in other sinks)
       */
      if (sink->display && !sink->window)       /* -> the window was toplevel */
        g_clear_object (&sink->display);

      g_mutex_unlock (&sink->display_lock);
      g_clear_object (&sink->pool);
      break;
    default:
      break;
  }

  return ret;
}

static void
gst_wayland_sink_set_context (GstElement * element, GstContext * context)
{
  GstWaylandSink *sink = GST_WAYLAND_SINK (element);

  if (gst_context_has_context_type (context,
          GST_WAYLAND_DISPLAY_HANDLE_CONTEXT_TYPE)) {
    g_mutex_lock (&sink->display_lock);
    if (G_LIKELY (!sink->display)) {
      gst_wayland_sink_set_display_from_context (sink, context);
    } else {
      GST_WARNING_OBJECT (element, "changing display handle is not supported");
      g_mutex_unlock (&sink->display_lock);
      return;
    }
    g_mutex_unlock (&sink->display_lock);
  }

  if (GST_ELEMENT_CLASS (parent_class)->set_context)
    GST_ELEMENT_CLASS (parent_class)->set_context (element, context);
}

static GstCaps *
gst_wayland_sink_get_caps (GstBaseSink * bsink, GstCaps * filter)
{
  GstWaylandSink *sink;
  GstCaps *caps;

  sink = GST_WAYLAND_SINK (bsink);

  caps = gst_pad_get_pad_template_caps (GST_VIDEO_SINK_PAD (sink));
  caps = gst_caps_make_writable (caps);

  g_mutex_lock (&sink->display_lock);

  if (sink->display) {
    GValue shm_list = G_VALUE_INIT, dmabuf_list = G_VALUE_INIT;
    GValue value = G_VALUE_INIT;
    GArray *formats, *modifiers;
    gint i;
    guint fmt;
    GstVideoFormat gfmt;
    guint64 mod;

    g_value_init (&shm_list, GST_TYPE_LIST);
    g_value_init (&dmabuf_list, GST_TYPE_LIST);

    /* Add corresponding shm formats */
    formats = sink->display->shm_formats;
    for (i = 0; i < formats->len; i++) {
      fmt = g_array_index (formats, uint32_t, i);
      gfmt = gst_wl_shm_format_to_video_format (fmt);
      if (gfmt != GST_VIDEO_FORMAT_UNKNOWN) {
        g_value_init (&value, G_TYPE_STRING);
        g_value_set_static_string (&value, gst_video_format_to_string (gfmt));
        gst_value_list_append_and_take_value (&shm_list, &value);
      }
    }

    gst_structure_take_value (gst_caps_get_structure (caps, 0), "format",
        &shm_list);

    /* Add corresponding dmabuf formats */
    formats = sink->display->dmabuf_formats;
    modifiers = sink->display->dmabuf_modifiers;
    for (i = 0; i < formats->len; i++) {
      fmt = g_array_index (formats, uint32_t, i);
      mod = g_array_index (modifiers, guint64, i);
      if (mod != DRM_FORMAT_MOD_LINEAR && mod != DRM_FORMAT_MOD_INVALID)
        continue;
      gfmt = gst_wl_dmabuf_format_to_video_format (fmt);
      if (gfmt != GST_VIDEO_FORMAT_UNKNOWN) {
        g_value_init (&value, G_TYPE_STRING);
        g_value_set_static_string (&value, gst_video_format_to_string (gfmt));
        gst_value_list_append_and_take_value (&dmabuf_list, &value);
      }
    }

    g_value_init (&value, G_TYPE_STRING);
    g_value_set_static_string (&value, "ENCODED");
    gst_value_list_append_and_take_value (&dmabuf_list, &value);

    gst_structure_take_value (gst_caps_get_structure (caps, 1), "format",
        &dmabuf_list);

    GST_DEBUG_OBJECT (sink, "display caps: %" GST_PTR_FORMAT, caps);
  }

  g_mutex_unlock (&sink->display_lock);

 /* add overlay meta feature when subcompositor is bound */
  if (1) {
    GstCaps *new_caps = gst_caps_copy (caps);
    gsize i, size;

    size = gst_caps_get_size (new_caps);
    for (i = 0; i < size; i++) {
      GstCapsFeatures *f = gst_caps_get_features (new_caps, i);
      gst_caps_features_add (f,
          GST_CAPS_FEATURE_META_GST_VIDEO_OVERLAY_COMPOSITION);
    }

    gst_caps_append (new_caps, caps);
    caps = new_caps;
  }

  if (filter) {
    GstCaps *intersection;

    intersection =
        gst_caps_intersect_full (filter, caps, GST_CAPS_INTERSECT_FIRST);
    gst_caps_unref (caps);
    caps = intersection;
  }

  return caps;
}

static GstBufferPool *
gst_wayland_create_pool (GstWaylandSink * sink, GstCaps * caps, gsize size)
{
  GstBufferPool *pool = NULL;
  GstStructure *structure;
  GstAllocator *alloc;

  pool = g_object_new (gst_wayland_pool_get_type (), NULL);

  structure = gst_buffer_pool_get_config (pool);
  gst_buffer_pool_config_set_params (structure, caps, size, 2, 0);

  alloc = gst_wl_shm_allocator_get ();
  gst_buffer_pool_config_set_allocator (structure, alloc, NULL);
  if (!gst_buffer_pool_set_config (pool, structure)) {
    g_object_unref (pool);
    pool = NULL;
  }
  g_object_unref (alloc);

  return pool;
}

static gboolean
gst_wayland_sink_set_caps (GstBaseSink * bsink, GstCaps * caps)
{
  GstWaylandSink *sink;
  gboolean use_dmabuf;
  GstVideoFormat format;
  guint64 modifier = DRM_FORMAT_MOD_LINEAR;
  GstCapsFeatures *f;

  sink = GST_WAYLAND_SINK (bsink);

  GST_DEBUG_OBJECT (sink, "set caps %" GST_PTR_FORMAT, caps);

  use_dmabuf = gst_caps_features_contains (gst_caps_get_features (caps, 0),
      GST_CAPS_FEATURE_MEMORY_DMABUF);

  gst_caps_ref (caps);

  sink->secure_mem = FALSE;

  /* extract the modifier and change the format to base format */
  if (use_dmabuf) {
    GstStructure *s;
    const gchar *format;

    s = gst_caps_get_structure (caps, 0);
    format = gst_structure_get_string (s, "format");

    if (!g_strcmp0 (format, "ENCODED")) {
      caps = gst_caps_make_writable (caps);
      s = gst_caps_get_structure (caps, 0);

      format = gst_structure_get_string (s, "drm-format");
      if (!format)
        goto invalid_format;

      gst_structure_set (s, "format", G_TYPE_STRING, format, NULL);
      gst_structure_get_uint64 (s, "drm-modifier", &modifier);
      gst_structure_get_boolean (s, "drm-secure", &sink->secure_mem);
      gst_structure_remove_fields (s, "drm-format", "drm-modifier",
          "drm-secure", NULL);
    }
  }

  /* extract info from caps */
  if (!gst_video_info_from_caps (&sink->video_info, caps))
    goto invalid_format;

  sink->has_video_mastering =
      gst_video_mastering_display_info_from_caps (&sink->video_mastering, caps);
  sink->has_video_cll =
      gst_video_content_light_level_from_caps (&sink->video_cll, caps);

  format = GST_VIDEO_INFO_FORMAT (&sink->video_info);
  sink->modifier = modifier;
  sink->video_info_changed = TRUE;

  /* create a new pool for the new caps */
  if (sink->pool)
    gst_object_unref (sink->pool);
  sink->pool = gst_wayland_create_pool (sink, caps, sink->video_info.size);

  /* validate the format base on the memory type. */
  if (use_dmabuf) {
    if (!gst_wl_display_check_format_for_dmabuf (sink->display, format,
            modifier))
      goto unsupported_format;
  } else if (!gst_wl_display_check_format_for_shm (sink->display, format,
          modifier)) {
    goto unsupported_format;
  }

  sink->use_dmabuf = use_dmabuf;

  f = gst_caps_get_features (caps, 0);
  sink->use_overlay = gst_caps_features_contains (f,
      GST_CAPS_FEATURE_META_GST_VIDEO_OVERLAY_COMPOSITION);

  gst_caps_unref (caps);

  return TRUE;

invalid_format:
  {
    gst_caps_unref (caps);
    GST_ERROR_OBJECT (sink,
        "Could not locate image format from caps %" GST_PTR_FORMAT, caps);
    return FALSE;
  }
unsupported_format:
  {
    gst_caps_unref (caps);
    GST_ERROR_OBJECT (sink, "Format %s is not available on the display",
        gst_video_format_to_string (format));
    return FALSE;
  }
}

static void
gst_wayland_sink_get_overlay_size (GstWaylandSink * sink, gint * width, gint * height)
{
  gint overlay_w;
  gint overlay_h;

  overlay_w = sink->video_info.width;
  overlay_h = sink->video_info.height;

  if (sink->window && sink->window->render_rectangle.w > 0 &&
      sink->window->render_rectangle.h > 0) {
    GstVideoRectangle src = { 0, };
    GstVideoRectangle dst = { 0, };
    GstVideoRectangle res;

    /* center video inside the render window */
    src.w = overlay_w;
    src.h = overlay_h;
    dst.w = sink->window->render_rectangle.w;
    dst.h = sink->window->render_rectangle.h;

    gst_video_sink_center_rect (src, dst, &res, TRUE);

    overlay_w = res.w;
    overlay_h = res.h;
  }

  if (overlay_w > MAX_OVERLAY_WIDTH || overlay_h > MAX_OVERLAY_HEIGHT) {
    gdouble window_aspect = (gdouble) overlay_w / (gdouble) overlay_h;
    gdouble max_aspect = (gdouble) MAX_OVERLAY_WIDTH /
        (gdouble) MAX_OVERLAY_HEIGHT;

    if (window_aspect >= max_aspect) {
      overlay_w = MAX_OVERLAY_WIDTH;
      overlay_h = MAX_OVERLAY_WIDTH / window_aspect;
    } else {
      overlay_w = MAX_OVERLAY_HEIGHT * window_aspect;
      overlay_h = MAX_OVERLAY_HEIGHT;
    }
  }

  if (width)
    *width = overlay_w;
  if (height)
    *height = overlay_h;
}

static gboolean
gst_wayland_sink_propose_allocation (GstBaseSink * bsink, GstQuery * query)
{
  GstWaylandSink *sink = GST_WAYLAND_SINK (bsink);
  GstCaps *caps;
  GstBufferPool *pool = NULL;
  gboolean need_pool;
  GstStructure *allocation_meta = NULL;
  GstAllocator *alloc;
  gint width, height;

  gst_query_parse_allocation (query, &caps, &need_pool);

  if (need_pool)
    pool = gst_wayland_create_pool (sink, caps, sink->video_info.size);

  gst_query_add_allocation_pool (query, pool, sink->video_info.size, 2, 0);
  if (pool)
    g_object_unref (pool);

  alloc = gst_wl_shm_allocator_get ();
  gst_query_add_allocation_param (query, alloc, NULL);
  gst_query_add_allocation_meta (query, GST_VIDEO_META_API_TYPE, NULL);
  g_object_unref (alloc);

  gst_wayland_sink_get_overlay_size (sink, &width, &height);
  if (width != sink->video_info.width || height != sink->video_info.height) {
    allocation_meta =
        gst_structure_new ("GstVideoOverlayCompositionMeta",
        "width", G_TYPE_UINT, width,
        "height", G_TYPE_UINT, height, NULL);
  }

  gst_query_add_allocation_meta (query,
      GST_VIDEO_OVERLAY_COMPOSITION_META_API_TYPE, allocation_meta);

  return TRUE;
}

static void
frame_redraw_callback (void *data, struct wl_callback *callback, uint32_t time)
{
  GstWaylandSink *sink = data;

  GST_LOG ("frame_redraw_cb");

  g_mutex_lock (&sink->render_lock);
  sink->redraw_pending = FALSE;

  if (sink->callback) {
    wl_callback_destroy (callback);
    sink->callback = NULL;
  }
  g_mutex_unlock (&sink->render_lock);
}

static const struct wl_callback_listener frame_callback_listener = {
  frame_redraw_callback
};

static void
ensure_overlay (GstWaylandSink * sink)
{
  GstCaps *overlay_caps;
  GstStructure *config;
  GstBufferPool *pool;
  gint width, height;

  if (sink->overlay_pool && !sink->check_pool)
    return;

  sink->check_pool = FALSE;
  gst_wayland_sink_get_overlay_size (sink, &width, &height);

  /* create pool for overlay */
  overlay_caps = gst_caps_new_simple ("video/x-raw",
      "format", G_TYPE_STRING, "BGRA",
      "width", G_TYPE_INT, width,
      "height", G_TYPE_INT, height,
      NULL);

  if ((pool = sink->overlay_pool)) {
    GstCaps *pcaps;

    GST_DEBUG_OBJECT (sink, "check existing overlay pool caps");
    config = gst_buffer_pool_get_config (pool);
    gst_buffer_pool_config_get_params (config, &pcaps, NULL, NULL, NULL);
    if (!gst_caps_is_equal (overlay_caps, pcaps)) {
      gst_object_unref (pool);
      pool = NULL;
    }

    gst_structure_free (config);
  }

  if (!pool) {
    gst_video_info_from_caps (&sink->overlay_info, overlay_caps);
    GST_DEBUG_OBJECT (sink, "create new overlay pool %dx%d",
        sink->overlay_info.width, sink->overlay_info.height);
    pool = gst_wayland_create_pool (sink, overlay_caps,
        sink->overlay_info.size);
  }

  sink->overlay_pool = pool;
  gst_caps_unref (overlay_caps);
}

static GstVideoOverlayComposition *
scale_overlay (GstWaylandSink * sink, GstVideoOverlayComposition * comp)
{
  GstVideoOverlayRectangle *r;
  guint video_width;
  guint video_height;
  guint render_width;
  guint render_height;
  gdouble xoffset = 0;
  gdouble yoffset = 0;
  gdouble scale;
  guint i = 0;

  video_width = sink->video_info.width;
  video_height = sink->video_info.height;
  render_width = sink->overlay_info.width;
  render_height = sink->overlay_info.height;

  if (render_width == video_width && render_width == video_height) {
    /* no scaling applied to video, do nothing */
    return comp;
  }

  if (video_width * render_height >= video_height * render_width) {
    /* black borders at top and bottom of output */
    scale = (gdouble) render_width / (gdouble) video_width;
    yoffset = (render_height - (video_height * scale)) / 2.0;
  } else {
    /* black borders at left and right of output */
    scale = (gdouble) render_height / (gdouble) video_height;
    xoffset = (render_width - (video_width * scale)) / 2.0;
  }

  GST_LOG_OBJECT (sink, "update overlay render size, vid %dx%d, out %dx%d, "
      "scale %.3lf, offset %+.3lf%+.3lf", video_width, video_height,
      render_width, render_height, scale, xoffset, yoffset);

  comp = gst_video_overlay_composition_make_writable (comp);

  while ((r = gst_video_overlay_composition_get_rectangle (comp, i++))) {
    gdouble x, y, w, h;

    gst_video_overlay_rectangle_get_render_geometry (r, &x, &y, &w, &h);
    gst_video_overlay_rectangle_set_render_geometry (r,
        x * scale + xoffset, y * scale + yoffset,
        ceil(w * scale), ceil(h * scale));
  }

  return comp;
}

static void
clear_overlay (GstWaylandSink * sink)
{
  if (sink->seqnum_valid && sink->spu_surface) {
    GST_DEBUG_OBJECT (sink, "disable overlay plane");
    wl_surface_attach (sink->spu_surface, NULL, 0, 0);
    wl_surface_commit (sink->spu_surface);
    sink->seqnum_valid = FALSE;
  }
}

static GstFlowReturn
render_overlay (GstWaylandSink * sink, const GstVideoInfo * vinfo,
    GstVideoOverlayCompositionMeta * cmeta)
{
  GstBufferPoolAcquireParams acquire;
  GstVideoOverlayComposition *comp;
  GstBuffer *buffer;
  GstWlBuffer *wlbuffer;
  GstVideoFrame frame;
  GstFlowReturn ret;
  guint seqnum;

  if (!cmeta) {
    /* No overlay metadata is attached, disable overlay plane */
    clear_overlay (sink);
    return GST_FLOW_OK;
  }

  seqnum = gst_video_overlay_composition_get_seqnum (cmeta->overlay);
  if (sink->seqnum_valid && seqnum == sink->last_seqnum) {
    /* No need to render the same overlay twice */
    return GST_FLOW_OK;
  }

  ensure_overlay (sink);

  gst_buffer_pool_set_active (sink->overlay_pool, TRUE);

  /* acquire buffer from pool, and wait for a buffer to be released if
   * needed */
  acquire.flags = GST_BUFFER_POOL_ACQUIRE_FLAG_DONTWAIT;
  while ((ret = gst_buffer_pool_acquire_buffer (sink->overlay_pool, &buffer,
          &acquire)) == GST_FLOW_EOS) {
    GST_LOG_OBJECT (sink, "waiting for overlay buffer release");
    //wl_display_roundtrip_queue (sink->display, sink->queue);
    GST_ERROR_OBJECT (sink, "NO SPU BUFFER");
  }

  if (ret != GST_FLOW_OK) {
    GST_ERROR_OBJECT (sink, "failed to acquire buffer for overlay");
    return ret;
  }

  GST_DEBUG_OBJECT (sink, "render overlay");

  comp = gst_video_overlay_composition_ref (cmeta->overlay);
  comp = scale_overlay (sink, comp);

  gst_video_frame_map (&frame, &sink->overlay_info, buffer, GST_MAP_READWRITE);
  for (int i = 0; i < GST_VIDEO_INFO_N_PLANES (&frame.info); i++)
    memset (frame.map[i].data, 0, frame.map[i].size);
  gst_video_overlay_composition_blend (comp, &frame);
  gst_video_frame_unmap (&frame);
  gst_video_overlay_composition_unref (comp);

  GST_DEBUG_OBJECT (sink, "render overlay done");

  wlbuffer = gst_buffer_get_wl_buffer (sink->display, buffer);

  /* attach a wl_buffer if there isn't one yet */
  if (G_UNLIKELY (!wlbuffer)) {
    struct wl_buffer *wbuf;
    GstMemory *mem;

    mem = gst_buffer_peek_memory (buffer, 0);
    wbuf = gst_wl_shm_memory_construct_wl_buffer (mem, sink->display,
        &sink->overlay_info);

    if (G_UNLIKELY (!wbuf))
      goto no_wl_buffer_shm;

    gst_buffer_add_wl_buffer (buffer, wbuf, sink->display);
    wlbuffer = gst_buffer_get_wl_buffer (sink->display, buffer);
  }

  gst_buffer_unref (buffer);

  if (!sink->spu_surface) {
    sink->spu_surface = wl_compositor_create_surface (sink->display->compositor);
    if (sink->display->viewporter) {
      sink->spu_viewport =
        wp_viewporter_get_viewport (sink->display->viewporter, sink->spu_surface);
    }
    sink->spu_subsurface =
      wl_subcompositor_get_subsurface (sink->display->subcompositor,
          sink->spu_surface, sink->window->video_surface_wrapper);
  }

  if (sink->spu_viewport) {
    gint width = vinfo->width;
    gint height = vinfo->height;

    if (sink->window &&
        sink->window->render_rectangle.w &&
        sink->window->render_rectangle.h) {
      GstVideoRectangle src = { 0, };
      GstVideoRectangle dst = { 0, };
      GstVideoRectangle res;

      /* center the video inside window */
      src.w = sink->window->video_width;
      src.h = sink->window->video_height;
      dst.w = sink->window->render_rectangle.w;
      dst.h = sink->window->render_rectangle.h;

      gst_video_sink_center_rect (src, dst, &res, TRUE);

      /* center the overlay inside video */
      src.w = sink->overlay_info.width;
      src.h = sink->overlay_info.height;
      dst.w = res.w;
      dst.h = res.h;

      gst_video_sink_center_rect (src, dst, &res, TRUE);

      width = res.w;
      height = res.h;

      wl_subsurface_set_position (sink->spu_subsurface, res.x, res.y);
    }
    wp_viewport_set_destination (sink->spu_viewport, width, height);
  }

  gst_wl_buffer_attach (wlbuffer, sink->spu_surface);
  wl_surface_damage (sink->spu_surface, 0, 0, INT_MAX, INT_MAX);
  wl_surface_commit (sink->spu_surface);

  sink->last_seqnum = seqnum;
  sink->seqnum_valid = TRUE;

done:
  return ret;

no_wl_buffer_shm:
  {
    GST_ERROR_OBJECT (sink, "could not create wl_buffer out of wl_shm memory");
    ret = GST_FLOW_ERROR;
    goto done;
  }
}

/* must be called with the render lock */
static void
render_last_buffer (GstWaylandSink * sink, gboolean redraw)
{
  GstWlBuffer *wlbuffer;
  const GstVideoInfo *info;
  struct wl_surface *surface;
  struct wl_callback *callback;
  gboolean changed;

  wlbuffer = gst_buffer_get_wl_buffer (sink->display, sink->last_buffer);
  surface = gst_wl_window_get_wl_surface (sink->window);
  info = &sink->last_buffer_info;
  changed = sink->video_info_changed;

  sink->redraw_pending = TRUE;
  callback = wl_surface_frame (surface);
  sink->callback = callback;
  wl_callback_add_listener (callback, &frame_callback_listener, sink);

  if (G_UNLIKELY (sink->video_info_changed && !redraw)) {
    sink->last_buffer_info = sink->video_info;
    sink->video_info_changed = FALSE;
  }

  if (sink->use_overlay) {
    render_overlay (sink, info,
        gst_buffer_get_video_overlay_composition_meta (sink->last_buffer));
  } else {
    clear_overlay (sink);
  }

  gst_wl_window_render (sink->window, wlbuffer, info, changed);
}

static void
on_window_closed (GstWlWindow * window, gpointer user_data)
{
  GstWaylandSink *sink = GST_WAYLAND_SINK (user_data);

  /* Handle window closure by posting an error on the bus */
  GST_ELEMENT_ERROR (sink, RESOURCE, NOT_FOUND,
      ("Output window was closed"), (NULL));
}

static void
on_window_resized (GstWlWindow * window, gpointer user_data)
{
  GstWaylandSink *sink = GST_WAYLAND_SINK (user_data);

  gst_pad_push_event (GST_BASE_SINK (sink)->sinkpad,
      gst_event_new_reconfigure ());

  sink->check_pool = TRUE;
}

static GstFlowReturn
gst_wayland_sink_show_frame (GstVideoSink * vsink, GstBuffer * buffer)
{
  GstWaylandSink *sink = GST_WAYLAND_SINK (vsink);
  GstBuffer *to_render;
  GstWlBuffer *wlbuffer;
  GstVideoMeta *vmeta;
  GstVideoFormat format;
  guint64 modifier;
  GstVideoInfo src_info;
  GstMemory *mem;
  struct wl_buffer *wbuf = NULL;

  GstFlowReturn ret = GST_FLOW_OK;

  g_mutex_lock (&sink->render_lock);

  GST_LOG_OBJECT (sink, "render buffer %p", buffer);

  if (G_UNLIKELY (!sink->window)) {
    /* ask for window handle. Unlock render_lock while doing that because
     * set_window_handle & friends will lock it in this context */
    g_mutex_unlock (&sink->render_lock);
    gst_video_overlay_prepare_window_handle (GST_VIDEO_OVERLAY (sink));
    g_mutex_lock (&sink->render_lock);

    if (!sink->window) {
      /* if we were not provided a window, create one ourselves */
      sink->window = gst_wl_window_new_toplevel (sink->display,
          &sink->video_info, sink->fullscreen, &sink->render_lock);
      g_signal_connect_object (sink->window, "closed",
          G_CALLBACK (on_window_closed), sink, 0);
      g_signal_connect_object (sink->window, "resized",
          G_CALLBACK (on_window_resized), sink, 0);
    }
  }

  /* drop buffers until we get a frame callback */
  if (sink->redraw_pending) {
    GST_LOG_OBJECT (sink, "buffer %p dropped (redraw pending)", buffer);
    goto done;
  }

  /* make sure that the application has called set_render_rectangle() */
  if (G_UNLIKELY (sink->window->render_rectangle.w == 0))
    goto no_window_size;

  wlbuffer = gst_buffer_get_wl_buffer (sink->display, buffer);

  if (G_LIKELY (wlbuffer && wlbuffer->display == sink->display)) {
    GST_LOG_OBJECT (sink, "buffer %p has a wl_buffer from our display, "
        "writing directly", buffer);
    to_render = buffer;
    goto render;
  }

  /* update video info from video meta */
  mem = gst_buffer_peek_memory (buffer, 0);

  src_info = sink->video_info;
  vmeta = gst_buffer_get_video_meta (buffer);
  if (vmeta) {
    gint i;

    src_info.width = vmeta->width;
    src_info.height = vmeta->height;

    for (i = 0; i < vmeta->n_planes; i++) {
      src_info.offset[i] = vmeta->offset[i];
      src_info.stride[i] = vmeta->stride[i];
    }
    src_info.size = gst_buffer_get_size (buffer);
  }

  GST_LOG_OBJECT (sink, "buffer %p does not have a wl_buffer from our "
      "display, creating it", buffer);

  format = GST_VIDEO_INFO_FORMAT (&src_info);
  modifier = sink->modifier;
  if (gst_wl_display_check_format_for_dmabuf (sink->display, format, modifier)) {
    guint i, nb_dmabuf = 0;

    for (i = 0; i < gst_buffer_n_memory (buffer); i++)
      if (gst_is_dmabuf_memory (gst_buffer_peek_memory (buffer, i)))
        nb_dmabuf++;

    if (nb_dmabuf == 1 && (nb_dmabuf == gst_buffer_n_memory (buffer)))
      wbuf = gst_wl_gbm_buffer_backend_construct_wl_buffer (buffer,
          sink->display, &src_info, sink->secure_mem, modifier,
          sink->has_video_mastering, &sink->video_mastering,
          sink->has_video_cll, &sink->video_cll);

    if (!wbuf && nb_dmabuf && (nb_dmabuf == gst_buffer_n_memory (buffer)))
      wbuf = gst_wl_linux_dmabuf_construct_wl_buffer (buffer, sink->display,
          &src_info, modifier);
  }

  if (!wbuf && gst_wl_display_check_format_for_shm (sink->display, format,
          modifier)) {
    if (gst_buffer_n_memory (buffer) == 1 && gst_is_fd_memory (mem))
      wbuf = gst_wl_shm_memory_construct_wl_buffer (mem, sink->display,
          &src_info);

    /* If nothing worked, copy into our internal pool */
    if (!wbuf) {
      GstVideoOverlayCompositionMeta *cmeta;
      GstVideoFrame src, dst;

      /* we don't know how to create a wl_buffer directly from the provided
       * memory, so we have to copy the data to shm memory that we know how
       * to handle... */

      GST_LOG_OBJECT (sink, "buffer %p cannot have a wl_buffer, "
          "copying to wl_shm memory", buffer);

      /* sink->pool always exists (created in set_caps), but it may not
       * be active if upstream is not using it */
      if (!gst_buffer_pool_is_active (sink->pool)) {
        GstStructure *config;
        GstCaps *caps;

        config = gst_buffer_pool_get_config (sink->pool);
        gst_buffer_pool_config_get_params (config, &caps, NULL, NULL, NULL);

        /* revert back to default strides and offsets */
        gst_video_info_from_caps (&sink->video_info, caps);
        gst_buffer_pool_config_set_params (config, caps, sink->video_info.size,
            2, 0);

        /* This is a video pool, it should not fail with basic settings */
        if (!gst_buffer_pool_set_config (sink->pool, config) ||
            !gst_buffer_pool_set_active (sink->pool, TRUE))
          goto activate_failed;
      }

      ret = gst_buffer_pool_acquire_buffer (sink->pool, &to_render, NULL);
      if (ret != GST_FLOW_OK)
        goto no_buffer;

      cmeta = gst_buffer_get_video_overlay_composition_meta (buffer);
      if (cmeta) {
        gst_buffer_add_video_overlay_composition_meta (to_render,
            cmeta->overlay);
      }

      wlbuffer = gst_buffer_get_wl_buffer (sink->display, to_render);

      /* attach a wl_buffer if there isn't one yet */
      if (G_UNLIKELY (!wlbuffer)) {
        mem = gst_buffer_peek_memory (to_render, 0);
        wbuf = gst_wl_shm_memory_construct_wl_buffer (mem, sink->display,
            &sink->video_info);

        if (G_UNLIKELY (!wbuf))
          goto no_wl_buffer_shm;

        wlbuffer = gst_buffer_add_wl_buffer (to_render, wbuf, sink->display);
      }

      if (!gst_video_frame_map (&dst, &sink->video_info, to_render,
              GST_MAP_WRITE))
        goto dst_map_failed;

      if (!gst_video_frame_map (&src, &src_info, buffer, GST_MAP_READ)) {
        gst_video_frame_unmap (&dst);
        goto src_map_failed;
      }

      gst_video_frame_copy (&dst, &src);

      gst_video_frame_unmap (&src);
      gst_video_frame_unmap (&dst);

      goto render;
    }
  }

  if (!wbuf)
    goto no_wl_buffer;

  wlbuffer = gst_buffer_add_wl_buffer (buffer, wbuf, sink->display);
  to_render = buffer;

render:
  /* drop double rendering */
  if (G_UNLIKELY (wlbuffer ==
          gst_buffer_get_wl_buffer (sink->display, sink->last_buffer))) {
    GST_LOG_OBJECT (sink, "Buffer already being rendered");
    goto done;
  }

  gst_buffer_replace (&sink->last_buffer, to_render);
  render_last_buffer (sink, FALSE);

  if (buffer != to_render)
    gst_buffer_unref (to_render);
  goto done;

no_window_size:
  {
    GST_ELEMENT_ERROR (sink, RESOURCE, WRITE,
        ("Window has no size set"),
        ("Make sure you set the size after calling set_window_handle"));
    ret = GST_FLOW_ERROR;
    goto done;
  }
no_buffer:
  {
    GST_WARNING_OBJECT (sink, "could not create buffer");
    goto done;
  }
no_wl_buffer_shm:
  {
    GST_ERROR_OBJECT (sink, "could not create wl_buffer out of wl_shm memory");
    ret = GST_FLOW_ERROR;
    goto done;
  }
no_wl_buffer:
  {
    GST_ERROR_OBJECT (sink, "buffer %p cannot have a wl_buffer", buffer);
    ret = GST_FLOW_ERROR;
    goto done;
  }
activate_failed:
  {
    GST_ERROR_OBJECT (sink, "failed to activate bufferpool.");
    ret = GST_FLOW_ERROR;
    goto done;
  }
src_map_failed:
  {
    GST_ELEMENT_ERROR (sink, RESOURCE, READ,
        ("Video memory can not be read from userspace."), (NULL));
    ret = GST_FLOW_ERROR;
    goto done;
  }
dst_map_failed:
  {
    GST_ELEMENT_ERROR (sink, RESOURCE, WRITE,
        ("Video memory can not be written from userspace."), (NULL));
    ret = GST_FLOW_ERROR;
    goto done;
  }
done:
  {
    g_mutex_unlock (&sink->render_lock);
    return ret;
  }
}

static void
gst_wayland_sink_videooverlay_init (GstVideoOverlayInterface * iface)
{
  iface->set_window_handle = gst_wayland_sink_set_window_handle;
  iface->set_render_rectangle = gst_wayland_sink_set_render_rectangle;
  iface->expose = gst_wayland_sink_expose;
}

static void
gst_wayland_sink_set_window_handle (GstVideoOverlay * overlay, guintptr handle)
{
  GstWaylandSink *sink = GST_WAYLAND_SINK (overlay);
  struct wl_surface *surface = (struct wl_surface *) handle;

  g_return_if_fail (sink != NULL);

  if (sink->window != NULL && sink->window->video_surface == surface)
    return;

  g_mutex_lock (&sink->render_lock);

  GST_DEBUG_OBJECT (sink, "Setting window handle %" GST_PTR_FORMAT,
      (void *) handle);

  g_clear_object (&sink->window);

  if (sink->callback) {
    wl_callback_destroy (sink->callback);
    sink->callback = NULL;
  }
  sink->redraw_pending = FALSE;

  if (handle) {
    if (G_LIKELY (gst_wayland_sink_find_display (sink))) {
      /* we cannot use our own display with an external window handle */
      if (G_UNLIKELY (sink->display->own_display)) {
        GST_ELEMENT_ERROR (sink, RESOURCE, OPEN_READ_WRITE,
            ("Application did not provide a wayland display handle"),
            ("waylandsink cannot use an externally-supplied surface without "
                "an externally-supplied display handle. Consider providing a "
                "display handle from your application with GstContext"));
      } else {
        sink->window = gst_wl_window_new_in_surface (sink->display, surface,
            &sink->render_lock);
      }
    } else {
      GST_ERROR_OBJECT (sink, "Failed to find display handle, "
          "ignoring window handle");
    }
  }

  g_mutex_unlock (&sink->render_lock);
}

static void
gst_wayland_sink_set_render_rectangle (GstVideoOverlay * overlay,
    gint x, gint y, gint w, gint h)
{
  GstWaylandSink *sink = GST_WAYLAND_SINK (overlay);

  g_return_if_fail (sink != NULL);

  g_mutex_lock (&sink->render_lock);
  if (!sink->window) {
    g_mutex_unlock (&sink->render_lock);
    GST_WARNING_OBJECT (sink,
        "set_render_rectangle called without window, ignoring");
    return;
  }

  GST_DEBUG_OBJECT (sink, "window geometry changed to (%d, %d) %d x %d",
      x, y, w, h);
  gst_wl_window_set_render_rectangle (sink->window, x, y, w, h);

  g_mutex_unlock (&sink->render_lock);
}

static void
gst_wayland_sink_expose (GstVideoOverlay * overlay)
{
  GstWaylandSink *sink = GST_WAYLAND_SINK (overlay);

  g_return_if_fail (sink != NULL);

  GST_DEBUG_OBJECT (sink, "expose");

  g_mutex_lock (&sink->render_lock);
  if (sink->last_buffer) {
    GST_DEBUG_OBJECT (sink, "redrawing last buffer");
    if (sink->callback) {
      wl_callback_destroy (sink->callback);
      sink->callback = NULL;
    }
    sink->redraw_pending = FALSE;
    render_last_buffer (sink, TRUE);
  }
  g_mutex_unlock (&sink->render_lock);
}

static void
gst_wayland_sink_waylandvideo_init (GstWaylandVideoInterface * iface)
{
  iface->begin_geometry_change = gst_wayland_sink_begin_geometry_change;
  iface->end_geometry_change = gst_wayland_sink_end_geometry_change;
}

static void
gst_wayland_sink_begin_geometry_change (GstWaylandVideo * video)
{
  GstWaylandSink *sink = GST_WAYLAND_SINK (video);
  g_return_if_fail (sink != NULL);

  g_mutex_lock (&sink->render_lock);
  if (!sink->window) {
    g_mutex_unlock (&sink->render_lock);
    GST_INFO_OBJECT (sink,
        "begin_geometry_change called without window, ignoring");
    return;
  }

  g_mutex_unlock (&sink->render_lock);
}

static void
gst_wayland_sink_end_geometry_change (GstWaylandVideo * video)
{
  GstWaylandSink *sink = GST_WAYLAND_SINK (video);
  g_return_if_fail (sink != NULL);

  g_mutex_lock (&sink->render_lock);
  if (!sink->window) {
    g_mutex_unlock (&sink->render_lock);
    GST_INFO_OBJECT (sink,
        "end_geometry_change called without window, ignoring");
    return;
  }

  g_mutex_unlock (&sink->render_lock);
}

static gboolean
plugin_init (GstPlugin * plugin)
{
  GST_DEBUG_CATEGORY_INIT (gstwayland_debug, "waylandsink", 0,
      " wayland video sink");

  gst_wl_shm_allocator_register ();

  return GST_ELEMENT_REGISTER (waylandsink, plugin);
}

GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    waylandsink,
    "Wayland Video Sink", plugin_init, VERSION, "LGPL", GST_PACKAGE_NAME,
    GST_PACKAGE_ORIGIN)
